@@ -38,6 +38,7 @@
 #include <freerdp/crypto/certificate.h>
 #include <freerdp/utils/ringbuffer.h>
 #include <freerdp/utils/smartcardlogon.h>
+#include <freerdp/utils/aad.h>
 
 #include "arm.h"
 #include "wst.h"
@@ -53,10 +54,7 @@
 #include "../utils.h"
 #include "../redirection.h"
 
-//#define WITH_AAD
-#ifdef WITH_AAD
-#include <cjson/cJSON.h>
-#endif
+#include <winpr/json.h>
 
 #include <string.h>
 
@@ -67,6 +65,7 @@ struct rdp_arm
 	HttpContext* http;
 
 	UINT32 gateway_retry;
+	wLog* log;
 };
 
 typedef struct rdp_arm rdpArm;
@@ -74,7 +73,7 @@ typedef struct rdp_arm rdpArm;
 #define TAG FREERDP_TAG("core.gateway.arm")
 
 #ifdef WITH_AAD
-static BOOL arm_tls_connect(rdpArm* arm, rdpTls* tls, int timeout)
+static BOOL arm_tls_connect(rdpArm* arm, rdpTls* tls, UINT32 timeout)
 {
 	WINPR_ASSERT(arm);
 	WINPR_ASSERT(tls);
@@ -98,7 +97,7 @@ static BOOL arm_tls_connect(rdpArm* arm, rdpTls* tls, int timeout)
 
 	sockfd = freerdp_tcp_connect(arm->context, peerHostname, peerPort, timeout);
 
-	WLog_DBG(TAG, "connecting to %s %d", peerHostname, peerPort);
+	WLog_Print(arm->log, WLOG_DEBUG, "connecting to %s %d", peerHostname, peerPort);
 	if (sockfd < 0)
 		return FALSE;
 
@@ -127,7 +126,7 @@ static BOOL arm_tls_connect(rdpArm* arm, rdpTls* tls, int timeout)
 
 	if (isProxyConnection)
 	{
-		if (!proxy_connect(settings, bufferedBio, proxyUsername, proxyPassword,
+		if (!proxy_connect(arm->context, bufferedBio, proxyUsername, proxyPassword,
 		                   freerdp_settings_get_string(settings, FreeRDP_GatewayHostname),
 		                   (UINT16)freerdp_settings_get_uint32(settings, FreeRDP_GatewayPort)))
 		{
@@ -143,7 +142,7 @@ static BOOL arm_tls_connect(rdpArm* arm, rdpTls* tls, int timeout)
 	}
 
 	tls->hostname = freerdp_settings_get_string(settings, FreeRDP_GatewayHostname);
-	tls->port = settings->GatewayPort;
+	tls->port = MIN(UINT16_MAX, WINPR_ASSERTING_INT_CAST(int32_t, settings->GatewayPort));
 	tls->isGatewayTransport = TRUE;
 	status = freerdp_tls_connect(tls, bufferedBio);
 	if (status < 1)
@@ -159,6 +158,29 @@ static BOOL arm_tls_connect(rdpArm* arm, rdpTls* tls, int timeout)
 	return (status >= 1);
 }
 
+static BOOL arm_fetch_wellknown(rdpArm* arm)
+{
+	WINPR_ASSERT(arm);
+	WINPR_ASSERT(arm->context);
+	WINPR_ASSERT(arm->context->rdp);
+
+	rdpRdp* rdp = arm->context->rdp;
+	if (rdp->wellknown)
+		return TRUE;
+
+	const char* base =
+	    freerdp_settings_get_string(arm->context->settings, FreeRDP_GatewayAzureActiveDirectory);
+	const BOOL useTenant =
+	    freerdp_settings_get_bool(arm->context->settings, FreeRDP_GatewayAvdUseTenantid);
+	const char* tenantid = "common";
+	if (useTenant)
+		tenantid =
+		    freerdp_settings_get_string(arm->context->settings, FreeRDP_GatewayAvdAadtenantid);
+
+	rdp->wellknown = freerdp_utils_aad_get_wellknown(arm->log, base, tenantid);
+	return rdp->wellknown ? TRUE : FALSE;
+}
+
 static wStream* arm_build_http_request(rdpArm* arm, const char* method,
                                        TRANSFER_ENCODING transferEncoding, const char* content_type,
                                        size_t content_length)
@@ -172,9 +194,7 @@ static wStream* arm_build_http_request(rdpArm* arm, const char* method,
 	WINPR_ASSERT(content_type);
 
 	WINPR_ASSERT(arm->context);
-
-	freerdp* instance = arm->context->instance;
-	WINPR_ASSERT(instance);
+	WINPR_ASSERT(arm->context->rdp);
 
 	uri = http_context_get_uri(arm->http);
 	request = http_request_new();
@@ -189,15 +209,19 @@ static wStream* arm_build_http_request(rdpArm* arm, const char* method,
 	{
 		char* token = NULL;
 
-		if (!instance->GetAccessToken)
+		pGetCommonAccessToken GetCommonAccessToken = freerdp_get_common_access_token(arm->context);
+		if (!GetCommonAccessToken)
 		{
-			WLog_ERR(TAG, "No authorization token provided");
+			WLog_Print(arm->log, WLOG_ERROR, "No authorization token provided");
 			goto out;
 		}
 
-		if (!instance->GetAccessToken(instance, ACCESS_TOKEN_TYPE_AVD, &token, 0))
+		if (!arm_fetch_wellknown(arm))
+			goto out;
+
+		if (!GetCommonAccessToken(arm->context, ACCESS_TOKEN_TYPE_AVD, &token, 0))
 		{
-			WLog_ERR(TAG, "Unable to obtain access token");
+			WLog_Print(arm->log, WLOG_ERROR, "Unable to obtain access token");
 			goto out;
 		}
 
@@ -242,9 +266,7 @@ static BOOL arm_send_http_request(rdpArm* arm, rdpTls* tls, const char* method,
 		return FALSE;
 
 	const size_t sz = Stream_Length(s);
-
-	if (sz <= INT_MAX)
-		status = freerdp_tls_write_all(tls, Stream_Buffer(s), (int)sz);
+	status = freerdp_tls_write_all(tls, Stream_Buffer(s), sz);
 
 	Stream_Free(s, TRUE);
 	if (status >= 0 && content_length > 0 && data)
@@ -272,8 +294,9 @@ static rdpArm* arm_new(rdpContext* context)
 	if (!arm)
 		goto fail;
 
+	arm->log = WLog_Get(TAG);
 	arm->context = context;
-	arm->tls = freerdp_tls_new(context->settings);
+	arm->tls = freerdp_tls_new(context);
 	if (!arm->tls)
 		goto fail;
 
@@ -296,10 +319,10 @@ static char* arm_create_request_json(rdpArm* arm)
 
 	WINPR_ASSERT(arm);
 
-	cJSON* json = cJSON_CreateObject();
+	WINPR_JSON* json = WINPR_JSON_CreateObject();
 	if (!json)
 		goto arm_create_cleanup;
-	cJSON_AddStringToObject(
+	WINPR_JSON_AddStringToObject(
 	    json, "application",
 	    freerdp_settings_get_string(arm->context->settings, FreeRDP_RemoteApplicationProgram));
 
@@ -313,31 +336,33 @@ static char* arm_create_request_json(rdpArm* arm)
 	    freerdp_settings_get_uint32(arm->context->settings, FreeRDP_LoadBalanceInfoLength);
 	memcpy(lbi, freerdp_settings_get_pointer(arm->context->settings, FreeRDP_LoadBalanceInfo), len);
 
-	cJSON_AddStringToObject(json, "loadBalanceInfo", lbi);
-	cJSON_AddNullToObject(json, "LogonToken");
-	cJSON_AddNullToObject(json, "gatewayLoadBalancerToken");
+	WINPR_JSON_AddStringToObject(json, "loadBalanceInfo", lbi);
+	WINPR_JSON_AddNullToObject(json, "LogonToken");
+	WINPR_JSON_AddNullToObject(json, "gatewayLoadBalancerToken");
 
-	message = cJSON_PrintUnformatted(json);
+	message = WINPR_JSON_PrintUnformatted(json);
 arm_create_cleanup:
 	if (json)
-		cJSON_Delete(json);
+		WINPR_JSON_Delete(json);
 	free(lbi);
 	return message;
 }
 
 /**
- * treats the redirectedAuthBlob
+ * @brief treats the redirectedAuthBlob
  *
  * sample pbInput:
+ * @code
  *  41004500530000004b44424d01000000200000006ee71b295810b3fd13799da3825d0efa3a628e8f4a6eda609ffa975408556546
  *  'A\x00E\x00S\x00\x00\x00KDBM\x01\x00\x00\x00
  * \x00\x00\x00n\xe7\x1b)X\x10\xb3\xfd\x13y\x9d\xa3\x82]\x0e\xfa:b\x8e\x8fJn\xda`\x9f\xfa\x97T\x08UeF'
+ * @endcode
  *
  * @param pbInput the raw auth blob (base64 and utf16 decoded)
  * @param cbInput size of pbInput
  * @return the corresponding WINPR_CIPHER_CTX if success, NULL otherwise
  */
-static WINPR_CIPHER_CTX* treatAuthBlob(const BYTE* pbInput, size_t cbInput)
+static WINPR_CIPHER_CTX* treatAuthBlob(wLog* log, const BYTE* pbInput, size_t cbInput)
 {
 	WINPR_CIPHER_CTX* ret = NULL;
 	char algoName[100] = { 0 };
@@ -346,36 +371,37 @@ static WINPR_CIPHER_CTX* treatAuthBlob(const BYTE* pbInput, size_t cbInput)
 	                                     sizeof(algoName));
 	if (algoSz <= 0)
 	{
-		WLog_ERR(TAG, "invalid algoName");
+		WLog_Print(log, WLOG_ERROR, "invalid algoName");
 		return NULL;
 	}
 
 	algoName[algoSz] = 0;
 	if (strcmp(algoName, "AES") != 0)
 	{
-		WLog_ERR(TAG, "only AES is supported for now");
+		WLog_Print(log, WLOG_ERROR, "only AES is supported for now");
 		return NULL;
 	}
 
-	cbInput -= (algoSz + 1) * sizeof(WCHAR);
+	cbInput -= WINPR_ASSERTING_INT_CAST(size_t, (algoSz + 1)) * sizeof(WCHAR);
 
 	if (cbInput < 12)
 	{
-		WLog_ERR(TAG, "invalid AuthBlob size");
+		WLog_Print(log, WLOG_ERROR, "invalid AuthBlob size");
 		return NULL;
 	}
 
 	/* BCRYPT_KEY_DATA_BLOB_HEADER */
 	wStream staticStream = { 0 };
-	wStream* s =
-	    Stream_StaticConstInit(&staticStream, pbInput + (algoSz + 1) * sizeof(WCHAR), cbInput);
+	wStream* s = Stream_StaticConstInit(
+	    &staticStream, pbInput + WINPR_ASSERTING_INT_CAST(size_t, (algoSz + 1)) * sizeof(WCHAR),
+	    cbInput);
 
 	UINT32 dwMagic = 0;
 	Stream_Read_UINT32(s, dwMagic);
 
 	if (dwMagic != BCRYPT_KEY_DATA_BLOB_MAGIC)
 	{
-		WLog_ERR(TAG, "unsupported authBlob type");
+		WLog_Print(log, WLOG_ERROR, "unsupported authBlob type");
 		return NULL;
 	}
 
@@ -383,8 +409,8 @@ static WINPR_CIPHER_CTX* treatAuthBlob(const BYTE* pbInput, size_t cbInput)
 	Stream_Read_UINT32(s, dwVersion);
 	if (dwVersion != BCRYPT_KEY_DATA_BLOB_VERSION1)
 	{
-		WLog_ERR(TAG, "unsupported authBlob version %d, expecting %d", dwVersion,
-		         BCRYPT_KEY_DATA_BLOB_VERSION1);
+		WLog_Print(log, WLOG_ERROR, "unsupported authBlob version %" PRIu32 ", expecting %d",
+		           dwVersion, BCRYPT_KEY_DATA_BLOB_VERSION1);
 		return NULL;
 	}
 
@@ -394,11 +420,11 @@ static WINPR_CIPHER_CTX* treatAuthBlob(const BYTE* pbInput, size_t cbInput)
 
 	if (cbKeyData > cbInput)
 	{
-		WLog_ERR(TAG, "invalid authBlob size");
+		WLog_Print(log, WLOG_ERROR, "invalid authBlob size");
 		return NULL;
 	}
 
-	int cipherType = 0;
+	WINPR_CIPHER_TYPE cipherType = 0;
 	switch (cbKeyData)
 	{
 		case 16:
@@ -411,20 +437,20 @@ static WINPR_CIPHER_CTX* treatAuthBlob(const BYTE* pbInput, size_t cbInput)
 			cipherType = WINPR_CIPHER_AES_256_CBC;
 			break;
 		default:
-			WLog_ERR(TAG, "invalid authBlob cipher size");
+			WLog_Print(log, WLOG_ERROR, "invalid authBlob cipher size");
 			return NULL;
 	}
 
-	ret = winpr_Cipher_New(cipherType, WINPR_ENCRYPT, Stream_Pointer(s), NULL);
+	ret = winpr_Cipher_NewEx(cipherType, WINPR_ENCRYPT, Stream_Pointer(s), cbKeyData, NULL, 0);
 	if (!ret)
 	{
-		WLog_ERR(TAG, "error creating cipher");
+		WLog_Print(log, WLOG_ERROR, "error creating cipher");
 		return NULL;
 	}
 
 	if (!winpr_Cipher_SetPadding(ret, TRUE))
 	{
-		WLog_ERR(TAG, "unable to enable padding on cipher");
+		WLog_Print(log, WLOG_ERROR, "unable to enable padding on cipher");
 		winpr_Cipher_Free(ret);
 		return NULL;
 	}
@@ -454,7 +480,7 @@ static BOOL arm_stringEncodeW(const BYTE* pin, size_t cbIn, BYTE** ppOut, size_t
 	return TRUE;
 }
 
-static BOOL arm_encodeRedirectPasswd(rdpSettings* settings, const rdpCertificate* cert,
+static BOOL arm_encodeRedirectPasswd(wLog* log, rdpSettings* settings, const rdpCertificate* cert,
                                      WINPR_CIPHER_CTX* cipher)
 {
 	BOOL ret = FALSE;
@@ -469,7 +495,7 @@ static BOOL arm_encodeRedirectPasswd(rdpSettings* settings, const rdpCertificate
 	WCHAR* wpasswd = freerdp_settings_get_string_as_utf16(settings, FreeRDP_Password, &wpasswdLen);
 	if (!wpasswd)
 	{
-		WLog_ERR(TAG, "error when converting password to UTF16");
+		WLog_Print(log, WLOG_ERROR, "error when converting password to UTF16");
 		return FALSE;
 	}
 
@@ -481,7 +507,7 @@ static BOOL arm_encodeRedirectPasswd(rdpSettings* settings, const rdpCertificate
 	    !winpr_Cipher_Update(cipher, wpasswd, wpasswdBytes, encryptedPass, &encryptedPassLen) ||
 	    !winpr_Cipher_Final(cipher, encryptedPass + encryptedPassLen, &finalLen))
 	{
-		WLog_ERR(TAG, "error when ciphering password");
+		WLog_Print(log, WLOG_ERROR, "error when ciphering password");
 		goto out;
 	}
 	encryptedPassLen += finalLen;
@@ -492,21 +518,21 @@ static BOOL arm_encodeRedirectPasswd(rdpSettings* settings, const rdpCertificate
 	if (!freerdp_certificate_publickey_encrypt(cert, encryptedPass, encryptedPassLen, &output,
 	                                           &output_length))
 	{
-		WLog_ERR(TAG, "unable to encrypt with the server's public key");
+		WLog_Print(log, WLOG_ERROR, "unable to encrypt with the server's public key");
 		goto out;
 	}
 
 	size_t finalOutputLen = 0;
 	if (!arm_stringEncodeW(output, output_length, &finalOutput, &finalOutputLen))
 	{
-		WLog_ERR(TAG, "unable to base64+utf16 final blob");
+		WLog_Print(log, WLOG_ERROR, "unable to base64+utf16 final blob");
 		goto out;
 	}
 
 	if (!freerdp_settings_set_pointer_len(settings, FreeRDP_RedirectionPassword, finalOutput,
 	                                      finalOutputLen))
 	{
-		WLog_ERR(TAG, "unable to set the redirection password in settings");
+		WLog_Print(log, WLOG_ERROR, "unable to set the redirection password in settings");
 		goto out;
 	}
 
@@ -529,17 +555,17 @@ out:
  *       base64.b64decode( base64.b64decode(input).decode('utf-16') )
  *  in python
  */
-static BOOL arm_pick_base64Utf16Field(const cJSON* json, const char* name, BYTE** poutput,
-                                      size_t* plen)
+static BOOL arm_pick_base64Utf16Field(wLog* log, const WINPR_JSON* json, const char* name,
+                                      BYTE** poutput, size_t* plen)
 {
 	*poutput = NULL;
 	*plen = 0;
 
-	const cJSON* node = cJSON_GetObjectItemCaseSensitive(json, name);
-	if (!node || !cJSON_IsString(node))
+	WINPR_JSON* node = WINPR_JSON_GetObjectItemCaseSensitive(json, name);
+	if (!node || !WINPR_JSON_IsString(node))
 		return TRUE;
 
-	char* nodeValue = cJSON_GetStringValue(node);
+	const char* nodeValue = WINPR_JSON_GetStringValue(node);
 	if (!nodeValue)
 		return TRUE;
 
@@ -548,7 +574,8 @@ static BOOL arm_pick_base64Utf16Field(const cJSON* json, const char* name, BYTE*
 	crypto_base64_decode(nodeValue, strlen(nodeValue), &output1, &len1);
 	if (!output1 || !len1)
 	{
-		WLog_ERR(TAG, "error when first unbase64 for %s", name);
+		WLog_Print(log, WLOG_ERROR, "error when first unbase64 for %s", name);
+		free(output1);
 		return FALSE;
 	}
 
@@ -557,7 +584,8 @@ static BOOL arm_pick_base64Utf16Field(const cJSON* json, const char* name, BYTE*
 	free(output1);
 	if (!output2 || !len2)
 	{
-		WLog_ERR(TAG, "error when decode('utf-16') for %s", name);
+		WLog_Print(log, WLOG_ERROR, "error when decode('utf-16') for %s", name);
+		free(output2);
 		return FALSE;
 	}
 
@@ -566,7 +594,8 @@ static BOOL arm_pick_base64Utf16Field(const cJSON* json, const char* name, BYTE*
 	free(output2);
 	if (!output || !*plen)
 	{
-		WLog_ERR(TAG, "error when second unbase64 for %s", name);
+		WLog_Print(log, WLOG_ERROR, "error when second unbase64 for %s", name);
+		free(output);
 		return FALSE;
 	}
 
@@ -594,30 +623,119 @@ static BOOL arm_pick_base64Utf16Field(const cJSON* json, const char* name, BYTE*
  *
  */
 
-static BOOL arm_treat_azureInstanceNetworkMetadata(const char* metadata, rdpSettings* settings)
+static size_t arm_parse_ipvx_count(WINPR_JSON* ipvX)
+{
+	WINPR_ASSERT(ipvX);
+	WINPR_JSON* ipAddress = WINPR_JSON_GetObjectItem(ipvX, "ipAddress");
+	if (!ipAddress || !WINPR_JSON_IsArray(ipAddress))
+		return 0;
+	return WINPR_JSON_GetArraySize(ipAddress);
+}
+
+static BOOL arm_parse_ipv6(rdpSettings* settings, WINPR_JSON* ipv6, size_t* pAddressIdx)
+{
+	WINPR_ASSERT(settings);
+	WINPR_ASSERT(ipv6);
+	WINPR_ASSERT(pAddressIdx);
+
+	if (!freerdp_settings_get_bool(settings, FreeRDP_IPv6Enabled))
+		return TRUE;
+
+	WINPR_JSON* ipAddress = WINPR_JSON_GetObjectItem(ipv6, "ipAddress");
+	if (!ipAddress || !WINPR_JSON_IsArray(ipAddress))
+		return TRUE;
+	const size_t naddresses = WINPR_JSON_GetArraySize(ipAddress);
+	for (size_t j = 0; j < naddresses; j++)
+	{
+		WINPR_JSON* adressN = WINPR_JSON_GetArrayItem(ipAddress, j);
+		if (!adressN || !WINPR_JSON_IsString(adressN))
+			continue;
+
+		const char* addr = WINPR_JSON_GetStringValue(adressN);
+		if (utils_str_is_empty(addr))
+			continue;
+		if (!freerdp_settings_set_pointer_array(settings, FreeRDP_TargetNetAddresses,
+		                                        (*pAddressIdx)++, addr))
+			return FALSE;
+	}
+	return TRUE;
+}
+
+static BOOL arm_parse_ipv4(rdpSettings* settings, WINPR_JSON* ipv4, size_t* pAddressIdx)
+{
+	WINPR_ASSERT(settings);
+	WINPR_ASSERT(ipv4);
+	WINPR_ASSERT(pAddressIdx);
+
+	WINPR_JSON* ipAddress = WINPR_JSON_GetObjectItem(ipv4, "ipAddress");
+	if (!ipAddress || !WINPR_JSON_IsArray(ipAddress))
+		return TRUE;
+
+	const size_t naddresses = WINPR_JSON_GetArraySize(ipAddress);
+	for (size_t j = 0; j < naddresses; j++)
+	{
+		WINPR_JSON* adressN = WINPR_JSON_GetArrayItem(ipAddress, j);
+		if (!adressN)
+			continue;
+
+		WINPR_JSON* publicIpNode = WINPR_JSON_GetObjectItem(adressN, "publicIpAddress");
+		if (publicIpNode && WINPR_JSON_IsString(publicIpNode))
+		{
+			const char* publicIp = WINPR_JSON_GetStringValue(publicIpNode);
+			if (!utils_str_is_empty(publicIp))
+			{
+				if (!freerdp_settings_set_pointer_array(settings, FreeRDP_TargetNetAddresses,
+				                                        (*pAddressIdx)++, publicIp))
+					return FALSE;
+			}
+		}
+
+		WINPR_JSON* privateIpNode = WINPR_JSON_GetObjectItem(adressN, "privateIpAddress");
+		if (privateIpNode && WINPR_JSON_IsString(privateIpNode))
+		{
+			const char* privateIp = WINPR_JSON_GetStringValue(privateIpNode);
+			if (!utils_str_is_empty(privateIp))
+			{
+				if (!freerdp_settings_set_pointer_array(settings, FreeRDP_TargetNetAddresses,
+				                                        (*pAddressIdx)++, privateIp))
+					return FALSE;
+			}
+		}
+	}
+	return TRUE;
+}
+
+static BOOL arm_treat_azureInstanceNetworkMetadata(wLog* log, const char* metadata,
+                                                   rdpSettings* settings)
 {
 	BOOL ret = FALSE;
-	cJSON* json = cJSON_Parse(metadata);
+
+	WINPR_ASSERT(settings);
+
+	if (!freerdp_target_net_adresses_reset(settings, 0))
+		return FALSE;
+
+	WINPR_JSON* json = WINPR_JSON_Parse(metadata);
 	if (!json)
 	{
-		WLog_ERR(TAG, "invalid azureInstanceNetworkMetadata");
+		WLog_Print(log, WLOG_ERROR, "invalid azureInstanceNetworkMetadata");
 		return FALSE;
 	}
 
-	const cJSON* iface = cJSON_GetObjectItem(json, "interface");
+	WINPR_JSON* iface = WINPR_JSON_GetObjectItem(json, "interface");
 	if (!iface)
 	{
 		ret = TRUE;
 		goto out;
 	}
 
-	if (!cJSON_IsArray(iface))
+	if (!WINPR_JSON_IsArray(iface))
 	{
-		WLog_ERR(TAG, "expecting interface to be an Array");
+		WLog_Print(log, WLOG_ERROR, "expecting interface to be an Array");
 		goto out;
 	}
 
-	int interfaceSz = cJSON_GetArraySize(iface);
+	size_t interfaceSz = WINPR_JSON_GetArraySize(iface);
 	if (interfaceSz == 0)
 	{
 		WLog_WARN(TAG, "no addresses in azure instance metadata");
@@ -625,65 +743,75 @@ static BOOL arm_treat_azureInstanceNetworkMetadata(const char* metadata, rdpSett
 		goto out;
 	}
 
-	for (int i = 0; i < interfaceSz; i++)
+	size_t count = 0;
+	for (size_t i = 0; i < interfaceSz; i++)
 	{
-		const cJSON* interN = cJSON_GetArrayItem(iface, i);
+		WINPR_JSON* interN = WINPR_JSON_GetArrayItem(iface, i);
 		if (!interN)
 			continue;
 
-		const cJSON* ipv4 = cJSON_GetObjectItem(interN, "ipv4");
-		if (!ipv4)
-			continue;
+		WINPR_JSON* ipv6 = WINPR_JSON_GetObjectItem(interN, "ipv6");
+		if (ipv6)
+			count += arm_parse_ipvx_count(ipv6);
 
-		const cJSON* ipAddress = cJSON_GetObjectItem(ipv4, "ipAddress");
-		if (!ipAddress || !cJSON_IsArray(ipAddress))
-			continue;
-
-		int naddresses = cJSON_GetArraySize(ipAddress);
-		for (int j = 0; j < naddresses; j++)
-		{
-			const cJSON* adressN = cJSON_GetArrayItem(ipAddress, j);
-			if (!adressN)
-				continue;
-
-			const cJSON* publicIpNode = cJSON_GetObjectItem(adressN, "publicIpAddress");
-			if (!publicIpNode || !cJSON_IsString(publicIpNode))
-				continue;
-
-			char* publicIp = cJSON_GetStringValue(publicIpNode);
-			if (publicIp && strlen(publicIp) &&
-			    freerdp_settings_set_string(settings, FreeRDP_RedirectionTargetFQDN, publicIp))
-			{
-				WLog_INFO(TAG, "connecting to publicIp %s", publicIp);
-				ret = TRUE;
-				goto out;
-			}
-		}
+		WINPR_JSON* ipv4 = WINPR_JSON_GetObjectItem(interN, "ipv4");
+		if (ipv4)
+			count += arm_parse_ipvx_count(ipv4);
 	}
 
-	ret = TRUE;
+	if (!freerdp_target_net_adresses_reset(settings, count))
+		return FALSE;
+
+	size_t addressIdx = 0;
+	for (size_t i = 0; i < interfaceSz; i++)
+	{
+		WINPR_JSON* interN = WINPR_JSON_GetArrayItem(iface, i);
+		if (!interN)
+			continue;
+
+		WINPR_JSON* ipv6 = WINPR_JSON_GetObjectItem(interN, "ipv6");
+		if (ipv6)
+		{
+			if (!arm_parse_ipv6(settings, ipv6, &addressIdx))
+				goto out;
+		}
+
+		WINPR_JSON* ipv4 = WINPR_JSON_GetObjectItem(interN, "ipv4");
+		if (ipv4)
+		{
+			if (!arm_parse_ipv4(settings, ipv4, &addressIdx))
+				goto out;
+		}
+	}
+	if (addressIdx > UINT32_MAX)
+		goto out;
+
+	if (!freerdp_settings_set_uint32(settings, FreeRDP_TargetNetAddressCount, (UINT32)addressIdx))
+		goto out;
+
+	ret = freerdp_settings_get_uint32(settings, FreeRDP_TargetNetAddressCount) > 0;
 
 out:
-	cJSON_Delete(json);
+	WINPR_JSON_Delete(json);
 	return ret;
 }
 
-static BOOL arm_fill_rdstls(rdpArm* arm, rdpSettings* settings, const cJSON* json)
+static BOOL arm_fill_rdstls(rdpArm* arm, rdpSettings* settings, const WINPR_JSON* json,
+                            const rdpCertificate* redirectedServerCert)
 {
 	BOOL ret = TRUE;
 	BYTE* cert = NULL;
 	BYTE* authBlob = NULL;
-	rdpCertificate* redirectedServerCert = NULL;
 
 	do
 	{
 		/* redirectedAuthGuid */
-		const cJSON* redirectedAuthGuidNode =
-		    cJSON_GetObjectItemCaseSensitive(json, "redirectedAuthGuid");
-		if (!redirectedAuthGuidNode || !cJSON_IsString(redirectedAuthGuidNode))
+		WINPR_JSON* redirectedAuthGuidNode =
+		    WINPR_JSON_GetObjectItemCaseSensitive(json, "redirectedAuthGuid");
+		if (!redirectedAuthGuidNode || !WINPR_JSON_IsString(redirectedAuthGuidNode))
 			break;
 
-		char* redirectedAuthGuid = cJSON_GetStringValue(redirectedAuthGuidNode);
+		const char* redirectedAuthGuid = WINPR_JSON_GetStringValue(redirectedAuthGuidNode);
 		if (!redirectedAuthGuid)
 			break;
 
@@ -694,38 +822,33 @@ static BOOL arm_fill_rdstls(rdpArm* arm, rdpSettings* settings, const cJSON* jso
 		const SSIZE_T wGUID_len = ConvertUtf8ToWChar(redirectedAuthGuid, wGUID, ARRAYSIZE(wGUID));
 		if (wGUID_len < 0)
 		{
-			WLog_ERR(TAG, "unable to allocate space for redirectedAuthGuid");
+			WLog_Print(arm->log, WLOG_ERROR, "unable to allocate space for redirectedAuthGuid");
 			ret = FALSE;
 			goto endOfFunction;
 		}
 
-		BOOL status = freerdp_settings_set_pointer_len(settings, FreeRDP_RedirectionGuid, wGUID,
-		                                               (wGUID_len + 1) * sizeof(WCHAR));
+		BOOL status = freerdp_settings_set_pointer_len(
+		    settings, FreeRDP_RedirectionGuid, wGUID,
+		    WINPR_ASSERTING_INT_CAST(size_t, (wGUID_len + 1)) * sizeof(WCHAR));
 		if (!status)
 		{
-			WLog_ERR(TAG, "unable to set RedirectionGuid");
+			WLog_Print(arm->log, WLOG_ERROR, "unable to set RedirectionGuid");
 			ret = FALSE;
 			goto endOfFunction;
 		}
 
-		/* redirectedServerCert */
-		size_t certLen = 0;
-		if (!arm_pick_base64Utf16Field(json, "redirectedServerCert", &cert, &certLen))
-			break;
-
-		if (!rdp_redirection_read_target_cert(&redirectedServerCert, cert, certLen))
-			break;
-
 		/* redirectedAuthBlob */
 		size_t authBlobLen = 0;
-		if (!arm_pick_base64Utf16Field(json, "redirectedAuthBlob", &authBlob, &authBlobLen))
+		if (!arm_pick_base64Utf16Field(arm->log, json, "redirectedAuthBlob", &authBlob,
+		                               &authBlobLen))
 			break;
 
-		WINPR_CIPHER_CTX* cipher = treatAuthBlob(authBlob, authBlobLen);
+		WINPR_CIPHER_CTX* cipher = treatAuthBlob(arm->log, authBlob, authBlobLen);
 		if (!cipher)
 			break;
 
-		const BOOL rerp = arm_encodeRedirectPasswd(settings, redirectedServerCert, cipher);
+		const BOOL rerp =
+		    arm_encodeRedirectPasswd(arm->log, settings, redirectedServerCert, cipher);
 		winpr_Cipher_Free(cipher);
 		if (!rerp)
 			break;
@@ -734,7 +857,6 @@ static BOOL arm_fill_rdstls(rdpArm* arm, rdpSettings* settings, const cJSON* jso
 	} while (FALSE);
 
 	free(cert);
-	freerdp_certificate_free(redirectedServerCert);
 	free(authBlob);
 
 endOfFunction:
@@ -747,48 +869,69 @@ static BOOL arm_fill_gateway_parameters(rdpArm* arm, const char* message, size_t
 	WINPR_ASSERT(arm->context);
 	WINPR_ASSERT(message);
 
-	cJSON* json = cJSON_ParseWithLength(message, len);
+	rdpCertificate* redirectedServerCert = NULL;
+	WINPR_JSON* json = WINPR_JSON_ParseWithLength(message, len);
 	BOOL status = FALSE;
 	if (!json)
 		return FALSE;
 
 	rdpSettings* settings = arm->context->settings;
-	const cJSON* gwurl = cJSON_GetObjectItemCaseSensitive(json, "gatewayLocation");
-	if (cJSON_IsString(gwurl) && (gwurl->valuestring != NULL))
+	WINPR_JSON* gwurl = WINPR_JSON_GetObjectItemCaseSensitive(json, "gatewayLocation");
+	const char* gwurlstr = WINPR_JSON_GetStringValue(gwurl);
+	if (gwurlstr != NULL)
 	{
-		WLog_DBG(TAG, "extracted target url %s", gwurl->valuestring);
-		if (!freerdp_settings_set_string(settings, FreeRDP_GatewayUrl, gwurl->valuestring))
-			status = FALSE;
-		else
-			status = TRUE;
+		WLog_Print(arm->log, WLOG_DEBUG, "extracted target url %s", gwurlstr);
+		if (!freerdp_settings_set_string(settings, FreeRDP_GatewayUrl, gwurlstr))
+			goto fail;
 	}
 
-	const cJSON* serverNameNode = cJSON_GetObjectItem(json, "redirectedServerName");
+	WINPR_JSON* serverNameNode = WINPR_JSON_GetObjectItem(json, "redirectedServerName");
 	if (serverNameNode)
 	{
-		char* serverName = cJSON_GetStringValue(serverNameNode);
+		const char* serverName = WINPR_JSON_GetStringValue(serverNameNode);
 		if (serverName)
-			status = freerdp_settings_set_string(settings, FreeRDP_ServerHostname, serverName);
+		{
+			if (!freerdp_settings_set_string(settings, FreeRDP_ServerHostname, serverName))
+				goto fail;
+		}
 	}
 
-	const cJSON* azureMeta = cJSON_GetObjectItem(json, "azureInstanceNetworkMetadata");
-	if (azureMeta && cJSON_IsString(azureMeta))
+	WINPR_JSON* azureMeta = WINPR_JSON_GetObjectItem(json, "azureInstanceNetworkMetadata");
+	if (azureMeta && WINPR_JSON_IsString(azureMeta))
 	{
-		if (!arm_treat_azureInstanceNetworkMetadata(cJSON_GetStringValue(azureMeta), settings))
+		if (!arm_treat_azureInstanceNetworkMetadata(arm->log, WINPR_JSON_GetStringValue(azureMeta),
+		                                            settings))
 		{
-			WLog_ERR(TAG, "error when treating azureInstanceNetworkMetadata");
+			WLog_Print(arm->log, WLOG_ERROR, "error when treating azureInstanceNetworkMetadata");
+			goto fail;
 		}
 	}
 
+	/* redirectedServerCert */
+	size_t certLen = 0;
+	BYTE* cert = NULL;
+	if (arm_pick_base64Utf16Field(arm->log, json, "redirectedServerCert", &cert, &certLen))
+	{
+		const BOOL rc = rdp_redirection_read_target_cert(&redirectedServerCert, cert, certLen);
+		free(cert);
+		if (!rc)
+			goto fail;
+		else if (!rdp_set_target_certificate(settings, redirectedServerCert))
+			goto fail;
+	}
+
 	if (freerdp_settings_get_string(settings, FreeRDP_Password))
 	{
 		/* note: we retrieve some more fields for RDSTLS only if we have a password provided by the
 		 * user, otherwise these are useless: we will not be able to do RDSTLS
 		 */
-		status = arm_fill_rdstls(arm, settings, json);
+		status = arm_fill_rdstls(arm, settings, json, redirectedServerCert);
 	}
-
-	cJSON_Delete(json);
+	else
+		status = TRUE;
+fail:
+	WINPR_JSON_Delete(json);
+	freerdp_certificate_free(redirectedServerCert);
 	return status;
 }
 
@@ -799,7 +942,7 @@ static BOOL arm_handle_request_ok(rdpArm* arm, const HttpResponse* response)
 	if (strnlen(msg, len + 1) > len)
 		return FALSE;
 
-	WLog_DBG(TAG, "Got HTTP Response data: %s", msg);
+	WLog_Print(arm->log, WLOG_DEBUG, "Got HTTP Response data: %s", msg);
 	return arm_fill_gateway_parameters(arm, msg, len);
 }
 
@@ -817,32 +960,36 @@ static BOOL arm_handle_bad_request(rdpArm* arm, const HttpResponse* response, BO
 	if (strnlen(msg, len + 1) > len)
 		return FALSE;
 
-	WLog_DBG(TAG, "Got HTTP Response data: %s", msg);
+	WLog_Print(arm->log, WLOG_DEBUG, "Got HTTP Response data: %s", msg);
 
-	cJSON* json = cJSON_ParseWithLength(msg, len);
+	WINPR_JSON* json = WINPR_JSON_ParseWithLength(msg, len);
 	if (json == NULL)
 	{
-		const char* error_ptr = cJSON_GetErrorPtr();
+		const char* error_ptr = WINPR_JSON_GetErrorPtr();
 		if (error_ptr != NULL)
-		{
-			WLog_ERR(TAG, "NullPoException: %s", error_ptr);
-			return FALSE;
-		}
+			WLog_Print(arm->log, WLOG_ERROR, "NullPoException: %s", error_ptr);
+
+		return FALSE;
 	}
 
-	const cJSON* gateway_code_str = cJSON_GetObjectItemCaseSensitive(json, "Code");
-	if (!cJSON_IsString(gateway_code_str) || (gateway_code_str->valuestring == NULL))
+	WINPR_JSON* gateway_code_obj = WINPR_JSON_GetObjectItemCaseSensitive(json, "Code");
+	const char* gw_code_str = WINPR_JSON_GetStringValue(gateway_code_obj);
+	if (gw_code_str == NULL)
 	{
-		WLog_ERR(TAG, "Response has no \"Code\" property");
+		WLog_Print(arm->log, WLOG_ERROR, "Response has no \"Code\" property");
 		http_response_log_error_status(WLog_Get(TAG), WLOG_ERROR, response);
 		goto fail;
 	}
 
-	if (strcmp(gateway_code_str->valuestring, "E_PROXY_ORCHESTRATION_LB_SESSIONHOST_DEALLOCATED") ==
-	    0)
+	if (strcmp(gw_code_str, "E_PROXY_ORCHESTRATION_LB_SESSIONHOST_DEALLOCATED") == 0)
 	{
 		*retry = TRUE;
-		WLog_DBG(TAG, "Starting your VM. It may take up to 5 minutes");
+		WINPR_JSON* message = WINPR_JSON_GetObjectItemCaseSensitive(json, "Message");
+		const char* msgstr = WINPR_JSON_GetStringValue(message);
+		if (!msgstr)
+			WLog_WARN(TAG, "Starting your VM. It may take up to 5 minutes");
+		else
+			WLog_WARN(TAG, "%s", msgstr);
 	}
 	else
 	{
@@ -852,7 +999,7 @@ static BOOL arm_handle_bad_request(rdpArm* arm, const HttpResponse* response, BO
 
 	rc = TRUE;
 fail:
-	cJSON_Delete(json);
+	WINPR_JSON_Delete(json);
 	return rc;
 }
 
@@ -860,6 +1007,12 @@ static BOOL arm_handle_request(rdpArm* arm, BOOL* retry, DWORD timeout)
 {
 	WINPR_ASSERT(retry);
 
+	if (!arm_fetch_wellknown(arm))
+	{
+		*retry = TRUE;
+		return FALSE;
+	}
+
 	*retry = FALSE;
 
 	char* message = NULL;
@@ -919,10 +1072,10 @@ arm_error:
 
 #endif
 
-BOOL arm_resolve_endpoint(rdpContext* context, DWORD timeout)
+BOOL arm_resolve_endpoint(wLog* log, rdpContext* context, DWORD timeout)
 {
 #ifndef WITH_AAD
-	WLog_ERR(TAG, "arm gateway support not compiled in");
+	WLog_Print(log, WLOG_ERROR, "arm gateway support not compiled in");
 	return FALSE;
 #else
 
@@ -935,7 +1088,7 @@ BOOL arm_resolve_endpoint(rdpContext* context, DWORD timeout)
 	if ((freerdp_settings_get_uint32(context->settings, FreeRDP_LoadBalanceInfoLength) == 0) ||
 	    (freerdp_settings_get_string(context->settings, FreeRDP_RemoteApplicationProgram) == NULL))
 	{
-		WLog_ERR(TAG, "loadBalanceInfo and RemoteApplicationProgram needed");
+		WLog_Print(log, WLOG_ERROR, "loadBalanceInfo and RemoteApplicationProgram needed");
 		return FALSE;
 	}
 
@@ -951,15 +1104,23 @@ BOOL arm_resolve_endpoint(rdpContext* context, DWORD timeout)
 		{
 			freerdp* instance = context->instance;
 			WINPR_ASSERT(instance);
-			const SSIZE_T delay = IFCALLRESULT(-1, instance->RetryDialog, instance, "arm-transport",
-			                                   arm->gateway_retry, arm);
+			SSIZE_T delay = IFCALLRESULT(-1, instance->RetryDialog, instance, "arm-transport",
+			                             arm->gateway_retry, arm);
 			arm->gateway_retry++;
 			if (delay <= 0)
 				break; /* error or no retry desired, abort loop */
 			else
 			{
-				WLog_DBG(TAG, "Delay for %" PRIdz "ms before next attempt", delay);
-				Sleep(delay);
+				WLog_Print(arm->log, WLOG_DEBUG, "Delay for %" PRIdz "ms before next attempt",
+				           delay);
+				while (delay > 0)
+				{
+					DWORD slp = (UINT32)delay;
+					if (delay > UINT32_MAX)
+						slp = UINT32_MAX;
+					Sleep(slp);
+					delay -= slp;
+				}
 			}
 		}
 		rc = arm_handle_request(arm, &retry, timeout);
