@@ -581,7 +581,9 @@ static BOOL process_files(wClipboard* clipboard, const char* data, UINT32 pSize,
 	if (strncmp(data, prefix, prefix_len) != 0)
 		return FALSE;
 	data += prefix_len;
-	pSize -= prefix_len;
+	if (pSize < prefix_len)
+		return FALSE;
+	pSize -= WINPR_ASSERTING_INT_CAST(uint32_t, prefix_len);
 
 	BOOL rc = FALSE;
 	char* copy = strndup(data, pSize);
@@ -592,10 +594,12 @@ static BOOL process_files(wClipboard* clipboard, const char* data, UINT32 pSize,
 	char* tok = strtok_s(copy, "\n", &endptr);
 	while (tok)
 	{
-		size_t tok_len = strnlen(tok, pSize);
+		const size_t tok_len = strnlen(tok, pSize);
 		if (!process_uri(clipboard, tok, tok_len))
 			goto fail;
-		pSize -= tok_len;
+		if (pSize < tok_len)
+			goto fail;
+		pSize -= WINPR_ASSERTING_INT_CAST(uint32_t, tok_len);
 		tok = strtok_s(NULL, "\n", &endptr);
 	}
 	rc = TRUE;
@@ -792,13 +796,14 @@ static void* convert_filedescriptors_to_file_list(wClipboard* clipboard, UINT32
 		previous_at = curName;
 		while ((stop_at = stop_at_special_chars(previous_at)) != NULL)
 		{
-			char* tmp =
-			    strndup(previous_at, WINPR_ASSERTING_INT_CAST(size_t, stop_at - previous_at));
+			const intptr_t diff = stop_at - previous_at;
+			if (diff < 0)
+				goto loop_fail;
+			char* tmp = strndup(previous_at, WINPR_ASSERTING_INT_CAST(size_t, diff));
 			if (!tmp)
 				goto loop_fail;
 
-			rc = _snprintf(&dst[pos], WINPR_ASSERTING_INT_CAST(size_t, stop_at - previous_at + 1),
-			               "%s", tmp);
+			rc = _snprintf(&dst[pos], WINPR_ASSERTING_INT_CAST(size_t, diff + 1), "%s", tmp);
 			free(tmp);
 			if (rc < 0)
 				goto loop_fail;
